local Addon, private = ...

setfenv(1, private)

-- Private methods
-- ============================================================================

local function matchTypes(style, nameString, typeString, newSig, result, fmt)
	local names = string.gmatch(nameString, "[%w_.]+")
	local types = string.gmatch(typeString, "[%w_./]+")
	local concat = { }
	local name = names()
	local type = types()
	while(name and type) do
		result[name] = result[name] or { }
		if(type == "nil") then
			concat[#concat + 1] = string.format(style.code["nil"], "nil")
		else
			for s in string.gsplit(type, "/") do
				result[name][s] = true
			end
			concat[#concat + 1] = string.format(fmt, name)
		end
		name = names()
		type = types()
	end
	newSig[#newSig + 1] = table.concat(concat, ", ")
end

local function makeSignatureClickable(page, sig, newSig)
	local copy
	if(string.prefix(sig, "Event.UI.")) then
		local left = string.find(sig, "(", 1, true)
		local right = string.find(sig, ")", left, true)
		local params = (string.sub(sig, left + 1, right))
		copy = string.format("self:EventAttach(%s, function(self, %s\nend, '')", (string.sub(sig, 1, left - 1)), params)
	elseif(string.prefix(sig, "Event.")) then
		local left = string.find(sig, "(", 1, true)
		local right = string.find(sig, ")", left, true)
		local params = (string.sub(sig, left + 1, right))
		if(#params > 1 and not string.prefix(params, "handle")) then
			params = "(handle, " .. params
		else
			params = "(" .. params
		end
		copy = string.format("Command.Event.Attach(%s, function%s \nend, '')", (string.sub(sig, 1, left - 1)), params)
	else
		copy = string.sub(sig, 1, (string.find(sig, ")", 1, true)))
	end
	return MakeHyperlinkHtml(tostring(page), "#!copy", copy, newSig)
end

local function processSignature(page, style, sig, results, params)
	local newSig = { }
	
	-- Return types and function name
	local assign = string.find(sig, "=")
	if(assign) then
		matchTypes(style, string.sub(sig, 1, assign), string.sub(sig, string.find(sig, "--", assign, true), string.find(sig, "<-", assign, true)), newSig, results, style.code.result)
		newSig[#newSig + 1] = " = "
		newSig[#newSig + 1] = string.match(sig, "=%s*([%w:_.]+)")
	else
		newSig[#newSig + 1] = string.match(sig, "^%s*([%w:_.]+)")
	end
	
	-- Parameter types
	newSig[#newSig + 1] = "("
	do
		local paren = string.find(sig, "(", 1, true)
		local tail = string.find(sig, assign and "<-" or "--", paren, true) or paren
		local types = string.sub(sig, tail, -1)
		matchTypes(style, string.sub(sig, paren, string.find(sig, ")", paren, true)), types, newSig, paren == tail and { } or params, style.code.parameter)
	end
	newSig[#newSig + 1] = ")"
	return makeSignatureClickable(page, sig, table.concat(newSig))
end

local function buildParameters(self, page, style, module, anchor, point, parameters, parameterTypes, results, resultTypes, html, row, typeFormatCallback, createFrame)
	local params = { }
	for name in pairs(parameters) do params[#params + 1] = name end
	table.sort(params)
	local rets = { }
	for name in pairs(results) do rets[#rets + 1] = name end
	table.sort(rets)
	
	local hasTypes = next(parameterTypes) or next(resultTypes)
	local owner = tostring(page)
	
	local padding = style.padding
	local nameWidth = 0
	local typeWidth = 0
	local nameFrames = { }
	local typeFrames = { }
	local descriptionFrames = { }
	local function type(types)
		local t = { }
		for name in pairs(types) do
			t[#t + 1] = typeFormatCallback(name)
			if(html) then
				t[#t] = MakeLinks(style, owner, module, t[#t])
			end
		end
		table.sort(t)
		if(html) then
			for i = 1, #t do
				t[i] = MakeLinks(style, owner, module, t[i])
			end
		end
		return "[" .. table.concat(t, "|") .. "]"
	end
	local function build(i, t, fmt, types, desc)
		local name = t[i]
		local type = type(types[name] or {})
		local nameFrame = createFrame("Text", "a", self, function(text)
			text:SetFontSize(style.p.fontSize)
			text:SetEffectGlow(style.p.glow)
			text:SetFontColor(ColorFromHtml(style.p.color))
		end)
		local descriptionFrame = createFrame("Text", "p", self, function(text)
			text:SetFontSize(style.p.fontSize)
			text:SetEffectGlow(style.p.glow)
			text:SetFontColor(ColorFromHtml(style.p.color))
			text:SetWordwrap(true)
		end)
		nameFrame:SetBackgroundColor(0.8, 0.8, 0.8, (row % 2 == 1) and 0.1 or 0)
		descriptionFrame:SetBackgroundColor(0.8, 0.8, 0.8, (row % 2 == 1) and 0.1 or 0)
		descriptionFrame:SetWordwrap(true)
		nameFrame:SetText(string.format(fmt, name), true)
		nameFrames[#nameFrames + 1] = nameFrame
		descriptionFrames[#descriptionFrames + 1] = descriptionFrame
		nameWidth = math.max(nameWidth, nameFrame:GetWidth())
		if(html) then
			descriptionFrame:SetText(MakeLinks(style, owner, module, desc[t[i]] or ""), html)
		else
			descriptionFrame:SetText(desc[t[i]] or "", html)
		end
		if(hasTypes) then
			local typeFrame = createFrame("Text", "", self)
			typeFrame:SetBackgroundColor(0.8, 0.8, 0.8, (row % 2 == 1) and 0.1 or 0)
			typeFrame:SetText(type, true)
			typeFrames[#typeFrames + 1] = typeFrame
			typeWidth = math.max(typeWidth, typeFrame:GetWidth())
		end
		row = row + 1
	end
	for i = 1, #params do build(i, params, style.code.parameter, parameterTypes, parameters) end
	for i = 1, #rets do build(i, rets, style.code.result, resultTypes, results) end
	for i = 1, #nameFrames do
		local nameFrame = nameFrames[i]
		local typeFrame
		local descriptionFrame = descriptionFrames[i]
		if(i == 1) then
			nameFrame:SetPoint("TOPLEFT", anchor, point .. "LEFT", padding + padding + padding, 0)
		else
			nameFrame:SetPoint("TOPLEFT", nameFrames[i - 1], "BOTTOMLEFT")
		end
		nameFrame:SetWidth(nameWidth)
		if(hasTypes) then
			typeFrame = typeFrames[i]
			typeFrame:SetPoint("TOPLEFT", nameFrame, "TOPRIGHT", padding, 0)
			typeFrame:SetWidth(typeWidth)
			descriptionFrame:SetPoint("TOPLEFT", typeFrame, "TOPRIGHT", padding, 0)
		else
			descriptionFrame:SetPoint("TOPLEFT", nameFrame, "TOPRIGHT", padding, 0)
		end
		if(i == 1) then
			descriptionFrame:SetPoint("TOPRIGHT", anchor, point .. "RIGHT")
		else
			descriptionFrame:SetPoint("TOPRIGHT", descriptionFrames[i - 1], "BOTTOMRIGHT")
		end
		nameFrame:SetPoint("BOTTOM", descriptionFrame, "BOTTOM")
		if(hasTypes) then
			typeFrame:SetPoint("BOTTOM", descriptionFrame, "BOTTOM")
		end
	end
	return descriptionFrames[#descriptionFrames]
end

-- Public methods
-- ============================================================================

Widgets = Widgets or { }

function Widgets.fn(page, style, module, signatures, parameters, results, html, typeFormatCallback, createFrame)
	local self = createFrame("Frame", "", page)
	local anchor
	
	local resultTypes = { }
	local paramTypes = { }
	local row = 1
	local sigs = { }
	for i = 1, #signatures do	
		local signature = createFrame("Text", "p", self, function(text)
			text:SetFontSize(style.p.fontSize)
			text:SetEffectGlow(style.p.glow)
			text:SetFontColor(ColorFromHtml(style.p.color))
			text:SetWordwrap(true)
		end)
		if(i == 1) then
			signature:SetPoint("TOPLEFT", self, "TOPLEFT")
			signature:SetPoint("TOPRIGHT", self, "TOPRIGHT")
		else
			signature:SetPoint("TOPLEFT", sigs[i - 1], "BOTTOMLEFT")
			signature:SetPoint("TOPRIGHT", sigs[i - 1], "BOTTOMRIGHT")
		end
		sigs[i] = signature
		signature:SetText(processSignature(page, style, signatures[i], resultTypes, paramTypes), true)
		signature:SetBackgroundColor(0.8, 0.8, 0.8, (i % 2 == 1) and 0.1 or 0)
		row = row + 1
	end
	
	local bottom = buildParameters(self, page, style, module, #sigs > 0 and sigs[#sigs] or self, #sigs > 0 and "BOTTOM" or "TOP", parameters, paramTypes, results, resultTypes, html, row, typeFormatCallback, createFrame)
	
	if(bottom or #signatures > 0) then
		anchor = createFrame("Frame", "", self)
		anchor:SetVisible(false)
		anchor:SetPoint("TOPLEFT", self, "TOPLEFT")
		anchor:SetPoint("TOPRIGHT", self, "TOPRIGHT")
		anchor:SetPoint("BOTTOM", bottom or sigs[#sigs], "BOTTOM")
	else
		self:SetHeight(1)
		anchor = self
	end
	
	return self, anchor, 0
end
